package fthfortran ;
###########################################################################
# Licence: 
# Copyright (C) 2000 Olivier Cessenat.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
###########################################################################
# A data structure for fortran 90 files:
#   It contains the modules and types setting, as well as the interfaces
#   and procedures definitions. Also contains global procedures.
#   Can also handle fortran 77 global files.
###########################################################################
use strict;
use Carp ;
use File::Basename ;
use vars qw($debug %hashrecursion %antirecursions @arrayrecursion $opt_oldmm) ;
$debug = 0 ;
###########################################################################
use fthmod ;
use fthtype ;
use fthinter ;
use fthrout;
use fthinp ;
use fthinc ;
use ihtml;
use chtml;
use oref;
###########################################################################
sub new {
  my $self = {} ;
  shift ;
  my(@args) = (@_) ;
  bless $self;
  if (@args) {
    $self->init(@args) ;
    $debug = $self->get(-debug) unless $debug ;
  }
  return $self ;
}
###########################################################################
sub setDebug {
  $debug = shift ;
}
###########################################################################
sub init {
  my $self = shift ;
  my(%args) = (@_) ;
  my($cle,$val) = () ;
  while(($cle,$val) = each %args) {
    if ($cle =~ /-(\w+)/) {
      $cle = $1 ;
      $self->{"$cle"} = $val ;
    }
  }
}
###########################################################################
sub get {
  my $self = shift ;
  my $param = shift ;
  if ($param =~ /-(\w+)/) {
    return($self->{"$1"}) ;
  } else {
    croak("Attempt to fetch a badly formed attribute") ;
    return(undef) ;
  }
}
###########################################################################
sub set {
  my $self = shift ;
  my $param = shift ;
  my $value = shift ;
  if ($param =~ /-(\w+)/) {
    $self->{"$1"} = $value ;
  } else {
    croak("Attempt to set a badly formed attribute") ;
    return(0) ;
  }
}
###########################################################################
# Dumps all the data to stdout
sub datadump {
  my $self = shift ;
  my $t = shift ;
  print STDOUT (caller(1))[3] . " $self name ".$self->{"name"}."={\n ";
  while(my($cle,$val) = each(%{$self})) {
    if (($val) && (ref($val) eq 'ARRAY')) {
      my(@array) = (@{$val}) ;
      print STDOUT "\t$cle = (" ;
      foreach my $sval (@array) {
	print STDOUT "$sval, " ;
      }
      print STDOUT ")\n" ;
    } elsif (($val) && (ref($val) eq 'HASH')) {
      my(%array) = (%{$val}) ;
      print STDOUT "$t$cle = {" ;
      foreach my $sval (keys %array) {
	if (exists $array{$sval} && ref($array{$sval}) eq 'fthmod') {
	  print STDOUT "$t\t$sval => "  ;
	  $array{$sval}->datadump("\t$t")  ;
	  print STDOUT", " ;
	} elsif (exists $array{$sval} && ref($array{$sval}) eq 'fthtype') {
	  print STDOUT "$t\t$sval => "  ;
	  $array{$sval}->datadump("\t$t")  ;
	  print STDOUT", " ;
	} elsif (exists $array{$sval} && ref($array{$sval}) eq 'fthinter') {
	  print STDOUT "$t\t$sval => "  ;
	  $array{$sval}->datadump("\t$t")  ;
	  print STDOUT", " ;
	} elsif (exists $array{$sval} && ref($array{$sval}) eq 'fthrout') {
	  print STDOUT "$t\t$sval => "  ;
	  $array{$sval}->datadump("\t$t")  ;
	  print STDOUT", " ;
	} elsif (exists $array{$sval} && ref($array{$sval}) eq 'fthinp') {
	  print STDOUT "$t\t$sval => "  ;
	  $array{$sval}->datadump("\t$t")  ;
	  print STDOUT", " ;
	} elsif (exists $array{$sval} && ref($array{$sval}) eq 'HASH') {
	  print STDOUT "$t\t$sval => {"  ;
	  $self->printHash("\t$t", $array{$sval}) ;
	  print STDOUT"}, " ;
	} else {
	  print STDOUT "$sval => $array{$sval}, " ;
	}
      }
      print STDOUT "},\n" ;
    } else {
      print STDOUT "\t$cle = $val\n" ;
    }
  }
  print "}\n" ;
}
###########################################################################
# Debug function.
sub debug {
  return unless $debug ;
  my $self = shift ;
  my $message = shift ;
  my ($a,$file,$b,$line0) = caller(0) ;
  my ($c,$fileb,$line1,$subr1) = caller(1) ;
  my ($d,$e,$line2,$subr2) = caller(2) ; 
  print STDOUT "TRACE FTHFORTRAN: file $file routine $subr1 called at line $line1 by $subr2 said\n" ;
  print STDOUT "VALUE FTHFORTRAN: $message\n" ;
}
###########################################################################
# Frees the data structure
sub free {
  my $self = shift ;
  foreach my $cle (keys %{$self}) {
    delete $self->{"$cle"} ;
  }
}
###########################################################################
# Adds a routine to a current stuff:
sub addRout {
  my $self = shift ;
  my $name = lc(shift) ;
  my $value = shift ;
  $self->{"FPROCEDURES"}->{"$name"} = $value ;
  print STDOUT "Adding a routine for $self, name=$name, object=$value\n" if $debug ;
}
###########################################################################
# Adds a type to a current stuff:
sub addType {
  my $self = shift ;
  my $name =lc(shift) ;
  my $value = shift ;
  $self->{"FTYPES"}->{"$name"} = $value ;
  print STDOUT "Adding a type for $self, name=$name, object=$value\n" if $debug ;
}
###########################################################################
# Adds a module to a current stuff:
sub addModule {
  my $self = shift ;
  my $name = lc(shift) ;
  my $value = shift ;
  $self->{"FMODULES"}->{"$name"} = $value ;
  print STDOUT "Adding a module for $self, name=$name, object=$value\n" if $debug ;
}
###########################################################################
# Adds a LaTeX input file
sub addInput {
  my $self = shift ;
  my $name = lc(shift) ;
  my $value = shift ;
  $self->{"INPUTS"}->{"$name"} = $value ;
  $value->addReferer($self) ;
  print STDOUT "Adding an input for $self(" . $self->get(-name) . "), name=$name, object=$value\n" if $debug ;
}
###########################################################################
# Adds an include:
sub addInclude {
  my $self = shift ;
  my $name = shift ;
  my $value = shift ;
  $self->{"INCLUDES"}->{"$name"} = $value ;
  $value->addReferer($self) ;
  print STDOUT "Adding a type for $self, name=$name, object=$value\n" if $debug ;
}
###########################################################################
sub getInputs {
  my $self = shift ;
  if (exists $self->{"INPUTS"}) {
    return(%{$self->{"INPUTS"}}) ;
  } else {
    return (()) ;
  }
}
###########################################################################
sub getIncludes {
  my $self = shift ;
  if (exists $self->{"INCLUDES"}) {
    return(%{$self->{"INCLUDES"}}) ;
  } else {
    return (()) ;
  }
}
###########################################################################
sub setDepends {
  my $self = shift ;
  if (exists $self->{"FPROCEDURES"}) {
    foreach my $name (keys %{$self->{"FPROCEDURES"}}) {
      # Current file defines $name
      my $orout = $self->{"FPROCEDURES"}->{"$name"} ;
      $orout->setDepends() ;
    }
  }
}
###########################################################################
sub revCalls {
  my $self = shift ;
  foreach my $name (keys %{$self->{"FPROCEDURES"}}) {
    # Current file defines $name
    my $orout = $self->{"FPROCEDURES"}->{"$name"} ;
    print STDOUT (caller(0))[3] . ": Procedure $name points to $orout, defined in file ".$self->get(-name) .":\n" if $debug ;
    $orout->revCalls() ;
  }
  foreach my $name (keys %{$self->{"FMODULES"}}) {
    my $omod = $self->{"FMODULES"}->{"$name"} ;
    print STDOUT (caller(0))[3] . ": Module $name points to $omod, defined in file ".$self->get(-name) .":\n" if $debug ;
    $omod->revCalls() ;
  }
}
###########################################################################
# Sets the local compile value:
sub setCompiles {
  my $self = shift ;
  my (%hash) = () ;
  # Gets the local dependency list:
  foreach my $name (keys %{$self->{"FMODULES"}}) {
    my $omod = $self->{"FMODULES"}->{"$name"} ;
    my (@array) = $omod->setCompiles() ;
    foreach my $a (@array) {
      $hash{$a} = 1 ;
    }
  }
  foreach my $name (keys %{$self->{"FPROCEDURES"}}) {
    my $orout = $self->{"FPROCEDURES"}->{"$name"} ;
    my (@array) = $orout->setCompiles() ;
    foreach my $a (@array) {
      $hash{$a} = 1 ;
    }
  }
  my $file = $self->get(-file) ;
  # Delete myself I need to be the last in array.
  delete $hash{$file} if (exists $hash{$file}) ;
  foreach my $file (keys %hash) {
    push(@{$self->{"COMPORDER"}}, $file) ;
  }
}
###########################################################################
# Fake setting for global:
sub setCompOrder {
  my $self = shift ;
  (@{$self->{"COMPORDER"}}) = (@_) ;
}
###########################################################################
# Gets all the direct files dependencies for current, make it once only
sub recCompiles {
  my $self = shift ;
  my (%objs) = (@_) ;
  my $file = $self->get(-file) ;
  print STDOUT "\t" . (caller(0))[3] . " Recursive call tests for $file for ".$self->get(-name)."\n" if $debug ;
  if ($file && $antirecursions{$file}) {
    if ($antirecursions{$file} == 1) {
      unless ($main::opt_v) {
	print STDOUT "\n" if $main::opt_show ;
      }
      print STDOUT "\t" . (caller(0))[3] . " Recursivity not allowed file $file for ".$self->get(-name)."\n" if $main::opt_show ;
      $antirecursions{$file} = 2 ;
    }
    return ;
  }
  $antirecursions{$file} = 1 if $file ;
  if (exists $self->{"COMPORDER"}) {
    foreach my $fic (@{$self->{"COMPORDER"}}) {
      unless (exists $hashrecursion{$fic}) {
	my $stuff = $objs{$fic} ;
	if ($stuff) {
	  $stuff->recCompiles(%objs) ;
	}
	unless (exists $hashrecursion{$fic}) {
	  push(@arrayrecursion, $fic) ;
	  $hashrecursion{$fic} = 1 ;
	}
      }
    }
  }
  if ($file) {
    unless (exists $hashrecursion{$file}) {
      $hashrecursion{$file} = 1 if $file ;
      push(@arrayrecursion, $file) ;
    }
  }
}
###########################################################################
# Gets all the direct files dependencies for current, make it once only
sub getCompiles {
  my $self = shift ;
  my (%objs) = (@_) ;
  if (!$self->get(-depends)) {
    local(%hashrecursion) = () ;
    local(@arrayrecursion) = () ;
    local(%antirecursions) = () ;
    $self->recCompiles(%objs) ;
    print STDOUT "\t" . (caller(0))[3] . " Arrayrecursion on (".$self->get(-name).") = @arrayrecursion\n" if $debug && $self->get(-name) ;
    (@{$self->{"COMPILELIST"}}) = (@arrayrecursion) ;
    $self->set(-depends => 1) ;
  }
  if (exists $self->{"COMPILELIST"}) {
    return(@{$self->{"COMPILELIST"}}) ;
  } else {
    return(()) ;
  }
}
###########################################################################
# Gets all the direct files dependencies for current, make it once only
sub oldCompiles {
  my $self = shift ;
  my (%objs) = (@_) ;
  print STDOUT (caller(1))[3] . ':' . __LINE__ . " In " . $self->get(-name) . " fetching calls($debug)\n" if $debug ;
  if (!$self->get(-depends)) {
    my (%hash) = () ;
    if (exists $self->{"COMPORDER"}) {
      foreach my $file (@{$self->{"COMPORDER"}}) {
	next unless $file ;
	unless (exists $hash{$file}) {
	  my $stuff = $objs{$file} ;
	  my $name = undef ;
	  $name = $stuff->get(-file) if $stuff ;
	  if ($name && $self->get(-name)) {
	    unless (!$stuff || exists $hash{$name} || $name eq $self->get(-name)) {
	      print STDOUT (caller(0))[3] . "Fetching sub call to ".$self->get(-name)." in file $file(@{$self->{COMPORDER}})\n" if $debug ;
	      my(@array) = $stuff->oldCompiles(%objs) ;
	      foreach my $a (@array) {
		unless (exists $hash{$a}) {
		  $hash{$a} = $a ;
		  push(@{$self->{"COMPILELIST"}}, $a) ;
		}
	      }
	    }
	  }
	}
	unless (exists $hash{$file}) {
	  $hash{$file} = $file ;
	  push(@{$self->{"COMPILELIST"}}, $file) ;
	}
      }
    }
    my $file = $self->get(-file) ;
    if ($file) {
      unless (exists $hash{$file}) {
	# Add myself:
	$hash{$file} = $file ;
	push(@{$self->{"COMPILELIST"}}, $file) ;
      }
    }
    $self->set(-depends => 1) ;
  }
  if (exists $self->{"COMPILELIST"}) {
    return(@{$self->{"COMPILELIST"}}) ;
  } else {
    return(()) ;
  }
}
###########################################################################
sub setCalls {
  my $self = shift ;
  print STDOUT (caller(0))[3] . ": Parsing calls for routs on ".$self->get(-name)."\n" if $debug;
  foreach my $name (keys %{$self->{"FPROCEDURES"}}) {
    # Current file defines $name
    my $orout = $self->{"FPROCEDURES"}->{"$name"} ;
    print STDOUT (caller(0))[3] . ": Procedure $name points to $orout, defined in file ".$self->get(-name) .":\n" if $debug ;
    $orout->setCalls() ;
  }
  print STDOUT (caller(0))[3] . ": Parsing calls for mods on ".$self->get(-name)."\n" if $debug;
  foreach my $name (keys %{$self->{"FMODULES"}}) {
    my $omod = $self->{"FMODULES"}->{"$name"} ;
    print STDOUT (caller(0))[3] . ":" . (caller(0))[2] . ": Module $name points to $omod, defined in file ".$self->get(-name) .":\n" if $debug ;
    $omod->setCalls() ;
  }
}
###########################################################################
sub sayCalls {
  my $self = shift ;
  my $ier = shift ;
  foreach my $name (keys %{$self->{"FPROCEDURES"}}) {
    my $orout = $self->{"FPROCEDURES"}->{"$name"} ;
    print STDOUT (caller(0))[3] . ": Routine $name points to $orout, defined in file ".$self->get(-name) .":\n" if $main::opt_v;
    $orout->sayCalls($ier) ;
  }
  foreach my $name (keys %{$self->{"FMODULES"}}) {
    my $omod = $self->{"FMODULES"}->{"$name"} ;
    print STDOUT (caller(0))[3] . ": Module $name points to $omod, defined in file ".$self->get(-name) .":\n" if $main::opt_v;
    $omod->sayCalls($ier) ;
  }
}
###########################################################################
# A routine for reading a fortran 77 file.
sub readf77 {
  my $self = shift ;
  my $file = shift ;
  my $masfic = shift ;
  my $in = shift ;
  my $oi = 0 ;
  my $chtml = &chtml::getCurrent() ;
  unless ($file) {
    $file = $self->get(-file) ;
    $masfic = $file ;
  }
  unless ((ref $in) && (fileno $in)) {
    croak("No API for $file") unless $file ;
    if (-f "$file") {
      if (open(FILE,"<$file")) {
	$chtml = new chtml(-file => $file, -name => $file) ;
	$in = \*FILE ;
	$oi = 1 ;
      }
    }
  }
  if ((ref $in) && (fileno $in)) {
    my $num ;
    my $preif = 0 ;
    my(@ploops) = () ;
    while(<$in>) {
      # Parsing a fortran 77 file below:
      chomp ;
      my $cc = "" ;
      if (/^(\#.*?)/) {
	# Some preprocessing stuff: passed to later analysis
      } elsif (/^(c)(.*)/i) {
	# Comment starter:
	$cc = $2 ;
	$_ = "" ;
      } elsif (/^(.*?)\!(.*)/) {
	$_ = $1 ;
	$cc = $2 ;
      }
      if (/^(\s*module\s+)(\w+)/i) {
	# Module reader below
	print STDOUT (caller(0))[3] . ':' . __LINE__ . ": Creating module $2\n" if $debug ;
	my $module = new fthmod(-name => "$2", -file => $file, -masfic => $masfic, -number => $.) ;
	$chtml->addDef($1) ;
	$chtml->addRaw($module) ;
	$self->addModule($2 => $module) ;
	$module->read($in) ;
      } elsif (/^(\s*type\s*\,*\s*.*\:\:\s*)(\w+)/i) {
	# Create a new type outside of a module: global type.
	print STDOUT (caller(0))[3] . ':' . __LINE__ . ": Creating type $2\n" if $debug ;
	my $type = new fthtype(-name => "$2", -file => $file, -masfic => $masfic, -global => 1, -number => $.) ;
	$chtml->addDef($1) ;
	$chtml->addRaw($type) ;
	$self->addType($2 => $type) ;
	$type->read($in) ;
      } elsif (/^(\s*type\s+)(\w+)/i) {
	# Create a new type outside of a module: global type.
	print STDOUT (caller(0))[3] . ':' . __LINE__ . ": Creating type $2\n" if $debug ;
	my $type = new fthtype(-name => "$2", -file => $file, -masfic => $masfic, -global => 1, -number => $.) ;
	$chtml->addDef($1) ;
	$chtml->addRaw($type) ;
	$self->addType($2 => $type) ;
	$type->read($in) ;
      } elsif (/^(\s*subroutine\s+)(\w+)(.*)/i) {
	# Create a new subroutine:
	print STDOUT (caller(0))[3] . ':' . __LINE__ . ": Creating routine $2\n" if $debug ;
	my $orout = new fthrout(-name => "$2", -file => $file, -masfic => $masfic, -global => $1, -number => $.) ;
	$chtml->addDef($1) ;
	$chtml->addRaw($orout) ;
	$self->addRout($2 => $orout) ;
	$orout->set(-theproto => 1) ;
	$orout->readf77subr($in, "     & $3") ;
      } elsif (/^(\s*program\s+)(\w+)/i) {
	# Create a new program: usually not inside a module !
	print STDOUT (caller(0))[3] . ':' . __LINE__ . ": Creating program $2\n" if $debug ;
	my $orout = new fthrout(-name => "$2", -file => $file, -masfic => $masfic, -global => $1, -number => $.) ;
	$chtml->addDef($1) ;
	$chtml->addRaw($orout) ;
	$self->addRout($2 => $orout) ;
	$orout->set(-theproto => 2) ;
	$orout->readf77subr($in, " ") ;
      } elsif (/^(\s*pure\s*function\s+)(\w+)(.*)/i) {
	# Create a new subroutine:
	print STDOUT (caller(0))[3] . ':' . __LINE__ . ": Creating function $1\n" if $debug ;
	my $orout = new fthrout(-name => "$2", -file => $file, -masfic => $masfic, -global => $1, -number => $.) ;
	$chtml->addDef($1) ;
	$chtml->addRaw($orout) ;
	$self->addRout($2 => $orout) ;
	$orout->readf77subr($in, "& $3") ;
	unless ($orout->get(-theproto)) {
	  $orout->addArgument($2) ;
	  $orout->addType($2 => 'void') ;
	  $orout->set(-theproto => 'void') ;
	}
      } elsif (/^(\s*)(.*?)\s*function\s+(\w+)(.*)/i) {
	# Create a new subroutine:
	print STDOUT (caller(0))[3] . ':' . __LINE__ . ": Creating function $3\n" if $debug ;
	my $orout = new fthrout(-name => "$3", -file => $file, -masfic => $masfic, -global => "function", -number => $.) ;
	$chtml->addDef($1.$2." function ") ;
	$chtml->addRaw($orout) ;
	$self->addRout($3 => $orout) ;
	$orout->set(-theproto => $2) if $2;
	$orout->readf77subr($in, "     & $4") ;
	if ($2) {
	  print STDOUT (caller(0))[3] . ':' . __LINE__ . ": Adding return type $2 to function $3\n" if $debug ;
	  $orout->addArgument($3) ;
	  $orout->addType($3 => $2) ;
	  $orout->setType($3 => $2) ;
	  $orout->set(-theproto => $2) ;
	  # $orout->datadump("\t") if $debug ;
	}
	unless ($orout->get(-theproto)) {
	  print STDOUT (caller(0))[3] . ':' . __LINE__ . ": Setting void proto to function $3\n" if $debug ;
	  $orout->addArgument($3) ;
	  $orout->addType($3 => 'void') ;
	  $orout->set(-theproto => 'void') ;
	}
      } elsif (/^\#(ifdef|ifndef|if|elif|else|endif)(.*)/i) {
	# Preprocessing tests, this seems good code below:
	if ($1 eq "endif") {
	  $num = pop(@ploops) ;
	  if ($num) {
	    $chtml->addRaw("<A Href=#C_$num>".'#'."$1</A>") ;
	    $preif-- ;
	  } else {
	    $chtml->addDef('#'."$1$2") ;
	  }
	} elsif ($1 eq "if" || $1 eq "ifdef" || $1 eq "ifndef") {
	  push(@ploops, $.) ;
	  $preif++ ;
	  $chtml->addRaw("<TEST><A Name=C_$.>".'#'."$1</A></TEST>$2") ;
	} else {
	  $num = pop(@ploops) ;
	  if ($num) {
	    push(@ploops, $num) ;
	    $chtml->addRaw("<A Href=#C_$num>".'#'."$1</A>") ;
	  } else {
	    $chtml->addDef('#'."$1$2") ;
	  }
	}
      } elsif (/^(\s*include\s+)([\"\'])([\w\.\/]+)\2(.*)/i || /^(\#include\s+)([\"\'])([\w\.\/]+)\2(.*)/i) {
	# Includes here:
	my $incf = $3 ;
	$chtml->addDef("$1$2") ;
	$chtml->addRaw("<A Name=C_$.></A>");
	unless (-f "$incf") {
	  my $base = dirname($self->get(-file)) ;
	  if (-f "$base/$incf") {
	    $incf = "$base/$incf" ;
	  } elsif (-f "$base/" . basename($incf)) {
	    $incf = "$base/" . basename($incf) ;
	  } elsif ($main::opt_I) {
	    # Possible...
	    my @TAGSHTML_PATH = (split ( /:/, $main::opt_I)) ;
	    foreach $base (@TAGSHTML_PATH) {
	      if (-f "$base/$incf") {
		$incf = "$base/$incf" ;
		last ;
	      }
	    }
	  }
	}
	if (-f "$incf") {
	  # No include policy for the moment:
	  my $fh ;
	  if (open($fh,"<$incf")) {
	    print STDOUT (caller(0))[3] . ": Open $incf from $file\n" if $debug ;
	    new chtml(-file => $incf, -name => $incf) ;
	    my $other = new fthinc(-file => "$incf", -masfic => $masfic, -name => "$incf") ;
	    $self->addInclude("$incf" => $other) ;
	    $chtml->addRaw($other) ;
	    $chtml->addDef("$2$4") ;
	    $self->readf77($incf, $masfic, $fh) ;
	    close($fh);
	    &chtml::setCurrent($chtml) ;
	    print STDOUT (caller(0))[3] . ": Closed $incf($.) from $file\n" if $debug ;
	  }
	} else {
	  my $refe = new oref(-name => "$3", -type => "fthinc") ;
	  $chtml->addRaw($refe) ;
	  $chtml->addDef("$2$4") ;
	}
      } else {
	$chtml->addDef($_) ;
      }
      if ($cc) {
	$chtml->addRaw('<COMMENT>') ;
	if ($cc =~ /(.*?\\)(reinput|verbatiminput|input)\s+([\w_\.\/]+)\.(tex|plt)(.*)/) {
	  $chtml->addLaCoM("<A Name=C_$.></A>") ;
	}
	while ($cc =~ /(.*?\\)(reinput|verbatiminput|input)\s+([\w_\.\/]+)\.(tex|plt)(.*)/) {
	  my $incp = "$3.$4" ;
	  $cc = $5 ;
	  $chtml->addLaTeX($1.$2." ") ;
#	  if ($incp =~ /^\// && ! -f  "$incp") {
#	    $incp = basename($incp) ;
#	  }
	  unless (-f "$incp") {
	    my $base = dirname($self->get(-file)) ;
	    my $nase = basename(dirname($incp)) ;
	    if (-f "$base/$incp") {
	      $incp = "$base/$incp" ;
	    } elsif (-f "$base/" . basename($incp)) {
	      $incp = "$base/" . basename($incp) ;
	    } elsif (-f "$nase/" . basename($incp)) {
	      $incp = "$nase/" . basename($incp) ;
	    } elsif ($main::opt_L) {
	      # LaTeX inputs directories
	      # Possible...
	      my @TAGSHTML_PATH = (split ( /:/, $main::opt_L)) ;
	      foreach $base (@TAGSHTML_PATH) {
		if (-f "$base/$incp") {
		  $incp = "$base/$incp" ;
		  last ;
		}
	      }
	    }
	  }
	  if (-f "$incp") {
	    my $other = new fthinp(-file => "$incp", -name => "$incp") ;
	    $self->addInput("$incp" => $other) ;
	    $chtml->addLaCoM($other) ;
	  } else {
	    my $refe = new oref(-name => $incp, -type => "fthinp") ;
	    $chtml->addLaCoM($refe) ;
	  }
	}
	$chtml->addLaTeX($cc) ;
	$chtml->addRaw("</COMMENT>") ;
      }
    }
  }
  close($in) if $oi ;
}
###########################################################################
# A routine for reading a fortran 90 file.
sub readf90 {
  my $self = shift ;
  my $file = shift ;
  my $masfic = shift ;
  my $in = shift ;
  my $oi = 0 ;
  my $chtml = &chtml::getCurrent() ;
  unless ($file) {
    $file = $self->get(-file) ;
    $masfic = $file ;
  }
  unless ((ref $in) && (fileno $in)) {
    croak("No API for $file") unless $file ;
    if (-f "$file") {
      if (open(FILE,"<$file")) {
	$chtml = new chtml(-file => $file, -name => $file) ;
	$in = \*FILE ;
	$oi = 1 ;
      }
    }
  }
  if ((ref $in) && (fileno $in)) {
    my $num ;
    my $preif = 0 ;
    my(@ploops) = () ;
    while(<$in>) {
      # Parsing a fortran 90 file below:
      chomp ;
      my $cc = "" ;
      if (/^(\#.*?)/) {
	# Some preprocessing stuff: passed to later analysis
      } elsif (/^(.*?)\!(.*)/) {
	$_ = $1 ;
	$cc = $2 ;
      }
      if (/^(\s*module\s+)(\w+)/i) {
	# Module reader below
	print STDOUT (caller(0))[3] . ':' . __LINE__ . ": Creating module $2\n" if $debug ;
	my $module = new fthmod(-name => "$2", -file => $file, -masfic => $masfic, -number => $.) ;
	$chtml->addDef($1) ;
	$chtml->addRaw($module) ;
	$self->addModule($2 => $module) ;
	$module->read($in) ;
      } elsif (/^(\s*type\s*\,*\s*.*\:\:\s*)(\w+)/i) {
	# Create a new type outside of a module: global type.
	print STDOUT (caller(0))[3] . ':' . __LINE__ . ": Creating type $2\n" if $debug ;
	my $type = new fthtype(-name => "$2", -file => $file, -masfic => $masfic, -global => 1, -number => $.) ;
	$chtml->addDef($1) ;
	$chtml->addRaw($type) ;
	$self->addType($2 => $type) ;
	$type->read($in) ;
      } elsif (/^(\s*type\s+)(\w+)/i) {
	# Create a new type outside of a module: global type.
	print STDOUT (caller(0))[3] . ':' . __LINE__ . ": Creating type $2\n" if $debug ;
	my $type = new fthtype(-name => "$2", -file => $file, -masfic => $masfic, -global => 1, -number => $.) ;
	$chtml->addDef($1) ;
	$chtml->addRaw($type) ;
	$self->addType($2 => $type) ;
	$type->read($in) ;
      } elsif (/^(\s*recursive\s+subroutine\s+)(\w+)(.*)/i) {
	# Create a new subroutine:
	print STDOUT (caller(0))[3] . ':' . __LINE__ . ": Creating routine $2\n" if $debug ;
	my $orout = new fthrout(-name => "$2", -file => $file, -masfic => $masfic, -global => $1, -number => $.) ;
	$chtml->addDef($1) ;
	$chtml->addRaw($orout) ;
	$self->addRout($2 => $orout) ;
	$orout->set(-theproto => 1) ;
	$orout->readf90subr($in, "& $3") ;
      } elsif (/^(\s*subroutine\s+)(\w+)(.*)/i) {
	# Create a new subroutine:
	print STDOUT (caller(0))[3] . ':' . __LINE__ . ": Creating routine $2\n" if $debug ;
	my $orout = new fthrout(-name => "$2", -file => $file, -masfic => $masfic, -global => $1, -number => $.) ;
	$chtml->addDef($1) ;
	$chtml->addRaw($orout) ;
	$self->addRout($2 => $orout) ;
	$orout->set(-theproto => 1) ;
	$orout->readf90subr($in, "& $3") ;
      } elsif (/^(\s*program\s+)(\w+)/i) {
	# Create a new program: usually not inside a module !
	print STDOUT (caller(0))[3] . ':' . __LINE__ . ": Creating program $2\n" if $debug ;
	my $orout = new fthrout(-name => "$2", -file => $file, -masfic => $masfic, -global => $1, -number => $.) ;
	$chtml->addDef($1) ;
	$chtml->addRaw($orout) ;
	$self->addRout($2 => $orout) ;
	$orout->set(-theproto => 2) ;
	$orout->readf90subr($in, " ") ;
      } elsif (/^(\s*pure\s*function\s+)(\w+)(.*)/i) {
	# Create a new subroutine:
	print STDOUT (caller(0))[3] . ':' . __LINE__ . ": Creating function $1\n" if $debug ;
	my $orout = new fthrout(-name => "$2", -file => $file, -masfic => $masfic, -global => $1, -number => $.) ;
	$chtml->addDef($1) ;
	$chtml->addRaw($orout) ;
	$self->addRout($2 => $orout) ;
	$orout->readf90subr($in, "& $3") ;
	unless ($orout->get(-theproto)) {
	  $orout->addArgument($2) ;
	  $orout->addType($2 => 'void') ;
	  $orout->set(-theproto => 'void') ;
	}
      } elsif (/^(\s*)(.*?)\s*function\s+(\w+)(.*)/i) {
	# Create a new subroutine:
	print STDOUT (caller(0))[3] . ':' . __LINE__ . ": Creating function $3\n" if $debug ;
	my $orout = new fthrout(-name => "$3", -file => $file, -masfic => $masfic, -global => "function", -number => $.) ;
	$chtml->addDef($1.$2." function ") ;
	$chtml->addRaw($orout) ;
	$self->addRout($3 => $orout) ;
	$orout->set(-theproto => $2) if $2;
	$orout->readf90subr($in, "& $4") ;
	if ($2) {
	  print STDOUT (caller(0))[3] . ':' . __LINE__ . ": Adding return type $2 to function $3\n" if $debug ;
	  $orout->addArgument($3) ;
	  $orout->addType($3 => $2) ;
	  $orout->setType($3 => $2) ;
	  $orout->set(-theproto => $2) ;
	  # $orout->datadump("\t") if $debug ;
	}
	unless ($orout->get(-theproto)) {
	  $orout->addArgument($3) ;
	  $orout->addType($3 => 'void') ;
	  $orout->set(-theproto => 'void') ;
	}
      } elsif (/^\#(ifdef|ifndef|if|elif|else|endif)(.*)/i) {
	# Preprocessing tests, this seems good code below:
	if ($1 eq "endif") {
	  $num = pop(@ploops) ;
	  if ($num) {
	    $chtml->addRaw("<A Href=#C_$num>".'#'."$1</A>") ;
	    $preif-- ;
	  } else {
	    $chtml->addDef('#'."$1$2") ;
	  }
	} elsif ($1 eq "if" || $1 eq "ifdef" || $1 eq "ifndef") {
	  push(@ploops, $.) ;
	  $preif++ ;
	  $chtml->addRaw("<TEST><A Name=C_$.>".'#'."$1</A></TEST>$2") ;
	} else {
	  $num = pop(@ploops) ;
	  if ($num) {
	    push(@ploops, $num) ;
	    $chtml->addRaw("<A Href=#C_$num>".'#'."$1</A>") ;
	  } else {
	    $chtml->addDef('#'."$1$2") ;
	  }
	}
      } elsif (/^(\s*include\s+)([\"\'])([\w\.\/]+)\2(.*)/i || /^(\#include\s+)([\"\'])([\w\.\/]+)\2(.*)/i) {
	# Includes here:
	my $incf = $3 ;
	$chtml->addDef("$1$2") ;
	$chtml->addRaw("<A Name=C_$.></A>");
	unless (-f "$incf") {
	  my $base = dirname($self->get(-file)) ;
	  if (-f "$base/$incf") {
	    $incf = "$base/$incf" ;
	  } elsif (-f "$base/" . basename($incf)) {
	    $incf = "$base/" . basename($incf) ;
	  } elsif ($main::opt_I) {
	    # Possible...
	    my @TAGSHTML_PATH = (split ( /:/, $main::opt_I)) ;
	    foreach $base (@TAGSHTML_PATH) {
	      if (-f "$base/$incf") {
		$incf = "$base/$incf" ;
		last ;
	      }
	    }
	  }
	}
	if (-f "$incf") {
	  my $fh ;
	  my $other ;
	  if (open($fh,"<$incf")) {
	    print STDOUT (caller(0))[3] . ": Open $incf from $file\n" if $debug ;
	    $other = &fthinc::getObject($incf) ;
	    if ($other) {
	      new chtml() ;
	    } else {
	      new chtml(-file => $incf, -name => $incf) ;
	      $other = new fthinc(-file => "$incf", -masfic => $masfic, -name => "$incf") ;
	    }
	    # my $f = new fthfortran(-file => $file, -name => $file) ;
	    $self->addInclude("$incf" => $other) ;
	    $chtml->addRaw($other) ;
	    $chtml->addDef("$2$4") ;
	    if (1) {
	      print STDOUT "Reading include $incf master=$masfic, handle=$fh\n" ;
	      $self->readf90($incf, $masfic, $fh) ;
	    } else {
	      # $f->readf90($incf, $masfic, $fh) ;
	    }
	    close($fh);
	    &chtml::setCurrent($chtml) ;
	    print STDOUT (caller(0))[3] . ": Closed $incf($.) from $file\n" if $debug ;
	  }
	} else {
	  my $refe = new oref(-name => "$3", -type => "fthinc") ;
	  $chtml->addRaw($refe) ;
	  $chtml->addDef("$2$4") ;
	}
      } else {
	$chtml->addDef($_) ;
      }
      if ($cc) {
	$chtml->addRaw('<COMMENT>') ;
	if ($cc =~ /(.*?\\)(reinput|verbatiminput|input)\s+([\w_\.\/]+)\.(tex|plt)(.*)/) {
	  $chtml->addLaCoM("<A Name=C_$.></A>") ;
	}
	while ($cc =~ /(.*?\\)(reinput|verbatiminput|input)\s+([\w_\.\/]+)\.(tex|plt)(.*)/) {
	  my $incp = "$3.$4" ;
	  $cc = $5 ;
	  $chtml->addLaTeX($1.$2." ") ;
#	  if ($incp =~ /^\// && ! -f  "$incp") {
#	    $incp = basename($incp) ;
#	  }
	  unless (-f "$incp") {
	    my $nase = basename(dirname($incp)) ;
	    my $base = dirname($self->get(-file)) ;
	    if (-f "$base/$incp") {
	      $incp = "$base/$incp" ;
	    } elsif (-f "$base/" . basename($incp)) {
	      $incp = "$base/" . basename($incp) ;
	    } elsif (-f "$nase/" . basename($incp)) {
	      $incp = "$nase/" . basename($incp) ;
	    } elsif ($main::opt_L) {
	      # LaTeX inputs directories
	      # Possible...
	      my @TAGSHTML_PATH = (split ( /:/, $main::opt_L)) ;
	      foreach $base (@TAGSHTML_PATH) {
		if (-f "$base/$incp") {
		  $incp = "$base/$incp" ;
		  last ;
		}
	      }
	    }
	  }
	  if (-f "$incp") {
	    my $other = new fthinp(-file => "$incp", -name => "$incp") ;
	    $self->addInput("$incp" => $other) ;
	    $chtml->addLaCoM($other) ;
	  } else {
	    my $refe = new oref(-name => $incp, -type => "fthinp") ;
	    $chtml->addLaCoM($refe) ;
	  }
	}
	$chtml->addLaTeX($cc) ;
	$chtml->addRaw("</COMMENT>") ;
      }
      # print STDOUT "$.:$_\n";
    }
  }
  close($in) if $oi ;
}
###########################################################################
# sort by line number:
sub bynumber {
 $a->{"number"} <=> $b->{"number"} ;
}
###########################################################################
# A routine for printing a fortran file:
sub htmlprint {
  my $self = shift ;
  my $outdir = shift ;
  my $reldir = shift ;
  my $file = $self->get(-file) ;
  my $on ;
  my $out ;
  my $outm ;
  my $outi ;
  my $outl ;
  my $outj ;
  my $masf ;
  $out = "$outdir/" . basename($file) ;
  print STDOUT (caller(0))[3] . ": Outfile($file) is $out\n" if $debug ;
  if ($out =~ /(.*)\.(.*)/) {
    # Implementation HTML file:
    $outm = $1 . '_' . $2 . '_c' . '.html' ;
    # Summary HTML file:
    $outi = $1 . '_' . $2 . '_i' . '.html' ;
    $outl = $1 . '_' . $2 . '_LF' . '.html' ;
    $outj = $1 . '_' . $2 . '.js' ;
    $masf = $1 . '_' . $2 . '.html' ;
  } else {
    $outm = $out . '_c' . '.html' ;
    $outi = $out . '_i' . '.html' ;
    $outl = $out . '_LF' . '.html' ;
    $outj = $out . '.js' ;
    $masf = $out . '.html' ;
  }
  my $ihtml ;
  if (open(MASTER,">$masf")) {
    my $handle = \*MASTER ;
    $ihtml = new ihtml(-name => $file, -handle => $handle, -file => $file, -reldir => $reldir) ;
    $ihtml->print_master() ;
    close(MASTER) ;
  }
  if (!$main::opt_nojava && open(OUFL,">$outl")) {
    my $handle = \*OUFL ;
    $ihtml->set(-handle => $handle) ;
    $ihtml->print_lf() ;
    close(OUFL) ;
  }
  if (!$main::opt_nojava && open(OUFJ,">$outj")) {
    my $ligne = "" ;
    my $js = \*OUFJ ;
    # Java tree building is initialized down there:
    $ligne .= $ihtml->start_tree() ."\n" ;
    $ligne .= $ihtml->addFolder("Descending tree", $file) . "\n" ;
    foreach my $orout (sort bynumber values %{$self->{"FPROCEDURES"}}) {
      $ligne .= $ihtml->irouts_down_tree($orout) ;
    }
    foreach my $omod (sort bynumber values %{$self->{"FMODULES"}}) {
      $ligne .= $ihtml->imods_down_tree($omod) ;
    }
    unless ($opt_oldmm) {
      $ligne .= "  ],\n" ;
    }
    $ligne .= $ihtml->addFolder("Ascending tree", $file) . "\n" ;
    foreach my $orout (sort bynumber values %{$self->{"FPROCEDURES"}}) {
      $ligne .= $ihtml->irouts_up_tree($orout) ;
    }
    foreach my $omod (sort bynumber values %{$self->{"FMODULES"}}) {
      $ligne .= $ihtml->imods_up_tree($omod) ;
    }
    unless ($opt_oldmm) {
      $ligne .= "  ],\n" ;
    }
    # Mark it is fully over:
    unless ($opt_oldmm) {
      $ligne .= "\t]\n];\n" ;
    }
    print $js $ligne ;
    close($js) ;
  }
  if (open(OUFI,">$outi")) {
    my $handle = \*OUFI ;
    my $title = "Interface for file $file" ;
    $ihtml->set(-handle => $handle) ;
    $ihtml->set(-title => $title) ;
    $ihtml->start_interface() ;
    my (@deps) = $self->getCompiles() ;
    if (@deps) {
      $ihtml->print_deps(@deps) ;
    }
    # Other stuffs:
    foreach my $otype (sort bynumber values %{$self->{"FTYPES"}}) {
      my $name = $otype->get(-name) ;
      # Printing the types:
      $ihtml->printi_type($otype) ;
    }
    foreach my $omod (sort bynumber values %{$self->{"FMODULES"}}) {
      my $name = $omod->get(-name) ;
      # Printing the modules:
      $ihtml->printi_mods($omod) ;
    }
    foreach my $orout (sort bynumber values %{$self->{"FPROCEDURES"}}) {
      my $name = $orout->get(-name) ;
      $ihtml->printi_routs($orout) ;
    }
    my(%hash) = $self->getIncludes() ;
    if (%hash) {
      print $handle "<BR>$file includes\n" ;
      foreach my $incname (sort keys %hash) {
	my $inco = $self->{"INCLUDES"}->{"$incname"} ;
	$ihtml->printi_incl($inco) ;
      }
    }
    if (exists $self->{"INPUTS"}) {
      print $handle "<BR>$file inputs\n" ;
      foreach my $inpname (sort keys %{$self->{"INPUTS"}}) {
	my $inpo = $self->{"INPUTS"}->{"$inpname"} ;
	$ihtml->printi_incp($inpo) ;
      }
    }
    $ihtml->end_html() ;
    close($handle) ;
  }
  if (open(OUFM,">$outm")) {
    my $handle = \*OUFM ;
    my $title = "Implementation of file $file" ;
    $ihtml->set(-title => $title) ;
    $ihtml->set(-handle => $handle) ;
    $ihtml->set(-js => 1) unless $main::opt_nojava;
    $ihtml->start_implementation() ;
    print $handle "<PRE>\n" ;
    # Get the chtml stuff:
    my $chtml = &chtml::getObject($file) ;
    $chtml->htmlprint($ihtml) ;
    print $handle "</PRE>\n" ;
    $ihtml->end_html() ;
    close($handle) ;
    $ihtml->free() ;
  }
}
###########################################################################
1;

